#ifndef SIMULATIONBOX_H
#define SIMULATIONBOX_H

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/RUMD_Error.h"
#include "rumd/rumd_base.h"
#include <cmath>

// The abstract Simulation Box class. Inherit this class to build your own SimulationBox.
class SimulationBox{
 
 private:
  SimulationBox(const SimulationBox&); 
  SimulationBox& operator=(const SimulationBox&); 
  
 protected:
  float* h_boxDetails;
  float* d_boxDetails;
  
 public:
  SimulationBox() : h_boxDetails(0), d_boxDetails(0){};
  virtual ~SimulationBox(){};
  
  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPointer );
  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float image1, float image2, float* simBoxPointer );
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPointer );
  __host__ __device__ void loadMemory( float* toMem, float* fromMem );

  virtual void ScaleBox( float factor ) = 0;
  virtual void CopyAnotherBox(const SimulationBox* boxToBeCopied) = 0;
  virtual SimulationBox* MakeCopy() = 0;

  // Set methods.
  virtual void SetBoxDetails( float* details ) = 0;

  // Get methods.
  virtual float GetVolume() const { return -1.f; };
  virtual float GetLength(int dir) const { return -1.f*dir; };
  virtual float4 GetSimulationBox() const { float4 value = {0,0,0,0}; return value; };

  virtual float* GetHostPointer(){
    if(!h_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Host pointer is invalid") );
    else
      return h_boxDetails;
  }

  virtual float* GetDevicePointer(){
    if(!d_boxDetails)
      throw( RUMD_Error("SimulationBox","SimulationBox","Device pointer is invalid") );
    else
      return d_boxDetails;
  }

  virtual void ScaleBoxDirection( float factor, int dir ) = 0;

};

///////////////////////////////////////////////////
// Avaliable simulation boxes
////////////////////////////////////////////////////

// A rectangular simulation box that assumes periodic boundaries.
class RectangularSimulationBox : public SimulationBox{

private:
  RectangularSimulationBox(const RectangularSimulationBox&);
  RectangularSimulationBox& operator=(const RectangularSimulationBox&);
  
  void AllocateBoxDetails();

protected:
  void CopyBoxDetailsToGPU();

 public:
  RectangularSimulationBox();
  RectangularSimulationBox( float Lx, float Ly, float Lz );
  ~RectangularSimulationBox();   

  void CopyAnotherBox(const SimulationBox* box_to_copy);
  SimulationBox* MakeCopy();
  
  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPointer ){
    float dr_x = position1.x - position2.x;
    float dr_y = position1.y - position2.y;
    float dr_z = position1.z - position2.z;

    if(periodicInX)
      dr_x -= simBoxPointer[1] * rintf( dr_x * simBoxPointer[4] );
    if(periodicInY)
      dr_y -= simBoxPointer[2] * rintf( dr_y * simBoxPointer[5] );
    if(periodicInZ)
      dr_z -= simBoxPointer[3] * rintf( dr_z * simBoxPointer[6] );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }

  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPointer ){
    float dr_x = (position1.x - position2.x) + simBoxPointer[1] * ( image1.x - image2.x );
    float dr_y = (position1.y - position2.y) + simBoxPointer[2] * ( image1.y - image2.y );
    float dr_z = (position1.z - position2.z) + simBoxPointer[3] * ( image1.z - image2.z );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPointer ){
    float4 image = { rintf( position.x * simBoxPointer[4] ), 
		     rintf( position.y * simBoxPointer[5] ), 
		     rintf( position.z * simBoxPointer[6] ), 
		     0.f };
    
    // Periodic boundary condition.
    if(periodicInX)
      position.x -= simBoxPointer[1] * image.x;
    if(periodicInY)
      position.y -= simBoxPointer[2] * image.y;
    if(periodicInZ)
      position.z -= simBoxPointer[3] * image.z;
    
    return image; 
  }
  
  __host__ __device__ void loadMemory( float* toMem, float* fromMem ){
    if((unsigned) fromMem[0]){
      toMem[1] = fromMem[1]; // boxLength
      toMem[2] = toMem[1]; 
      toMem[3] = toMem[1];
      toMem[4] = 1.f / toMem[1]; // 1 / boxLength
      toMem[5] = toMem[4];
      toMem[6] = toMem[4];
    }
    else{
      for(unsigned int i=0; i < simulationBoxSize; i++)
	toMem[i] = fromMem[i];
    }
  }
  
  void ScaleBox( float factor );
  void ScaleBoxDirection( float factor, int dir );
  
  // Set methods
  void SetBoxDetails( float* details );
  
  // Get methods
  float GetVolume() const { return h_boxDetails[1] * h_boxDetails[2] * h_boxDetails[3]; }
  float GetLength(int dir) const { return h_boxDetails[dir+1]; }

  float4 GetSimulationBox() const { float4 value = { h_boxDetails[1], h_boxDetails[2], h_boxDetails[3], GetVolume() };  return value; }

};



////////////////////////////////////////////////////////////////

// A rectangular simulation box that assumes periodic boundaries but with 
// Lees-Edwards boundary conditions to allow relative shear of the box and
// its images.

class LeesEdwardsSimulationBox : public RectangularSimulationBox{

 private:
  LeesEdwardsSimulationBox(const LeesEdwardsSimulationBox&);
  LeesEdwardsSimulationBox& operator=(const LeesEdwardsSimulationBox&);
  
  
 public:
  LeesEdwardsSimulationBox();
  LeesEdwardsSimulationBox( float Lx, float Ly, float Lz, float boxShift );
  LeesEdwardsSimulationBox(const SimulationBox* box_to_copy);
  ~LeesEdwardsSimulationBox() {}

  SimulationBox* MakeCopy();
  void CopyAnotherBox(const SimulationBox* box_to_copy);

  __host__ __device__ float4 calculateDistance( float4 position1, float4 position2, float* simBoxPointer ){
    float dr_x = position1.x - position2.x;
    float dr_y = position1.y - position2.y;
    float dr_z = position1.z - position2.z;

    float y_wrap = rintf( dr_y * simBoxPointer[5] );

    dr_y -= simBoxPointer[2] * y_wrap;
    dr_x -= y_wrap * simBoxPointer[7];

    dr_x -= simBoxPointer[1] * rintf( dr_x * simBoxPointer[4] );
    dr_z -= simBoxPointer[3] * rintf( dr_z * simBoxPointer[6] );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }

  __host__ __device__ float4 calculateDistanceWithImages( float4 position1, float4 position2, float4 image1, float4 image2, float* simBoxPointer ){
    float dr_x = (position1.x - position2.x) + simBoxPointer[1] * ( image1.x - image2.x );
    float dr_y = (position1.y - position2.y) + simBoxPointer[2] * ( image1.y - image2.y );
    float dr_z = (position1.z - position2.z) + simBoxPointer[3] * ( image1.z - image2.z );
    
    float4 value = { dr_x, dr_y, dr_z, dr_x * dr_x + dr_y * dr_y + dr_z * dr_z }; 
    return value;
  }
  
  __host__ __device__ float4 applyBoundaryCondition( float4& position, float* simBoxPointer ){
    float y_wrap = rintf( position.y * simBoxPointer[5] );
    position.x -= simBoxPointer[7] * y_wrap;

    float4 image = { rintf( position.x * simBoxPointer[4] ), 
		     y_wrap,
		     rintf( position.z * simBoxPointer[6] ), 
		     0.f };
    
    // Periodic boundary condition.
    position.x -= simBoxPointer[1] * image.x;
    position.y -= simBoxPointer[2] * image.y;
    position.z -= simBoxPointer[3] * image.z;
    
    return image; 
  }
  
  __host__ __device__ void loadMemory( float* toMem, float* fromMem ){
    if((unsigned) fromMem[0]){
      toMem[1] = fromMem[1]; // boxLength
      toMem[2] = toMem[1]; 
      toMem[3] = toMem[1];
      toMem[4] = 1.f / toMem[1]; // 1 / boxLength
      toMem[5] = toMem[4];
      toMem[6] = toMem[4];
      toMem[7] = fromMem[7];
    }
    else{
      for(unsigned int i=0; i < simulationBoxSize; i++)
	toMem[i] = fromMem[i];
    }
  }
  
  void ScaleBox( float factor );

  void IncrementBoxStrain(float inc_boxStrain);
  
  // Set methods
  void SetBoxShift(float set_boxShift);
  void ScaleBoxDirection( float factor, int dir );


  // Get methods
  float GetBoxShift() const { return h_boxDetails[7]; }
  float GetLength(int dir) const { return h_boxDetails[dir+1]; }

};

#endif // SIMULATIONBOX_H
