

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/AnglePotential.h"
#include "rumd/MoleculeData.h"
#include "rumd/Sample.h"



AngleCosSq::AngleCosSq(MoleculeData *globalM, Sample *globalS){

  M = globalM;
  S = globalS;
  SetID_String("AngleCosSq");

  num_threads = 32;
  num_blocks = M->num_angles/num_threads + 1;

}


__device__ inline void AngleAtomicFloatAdd(float *address, float val){
  int i_val = __float_as_int(val);
  int tmp0 = 0;
  int tmp1;
  
  while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0){
    tmp0 = tmp1;
    i_val = __float_as_int(val + __int_as_float(tmp1));
  }
}

__global__ void AngleResetForces( float4 *f, float4 *w ){

  f[blockIdx.x].x = f[blockIdx.x].y = f[blockIdx.x].z = f[blockIdx.x].w = 0.0;
  
  w[blockIdx.x].w = 0.0;

}


void AngleCosSq::CalcF(bool initialize, bool calc_stresses){


  unsigned num_angles = M->GetNumberOfAngles();
 
  if ( initialize )
    AngleResetForces<<<S->GetNumberOfParticles(),1>>>(S->particleData.d_f, 
						      S->particleData.d_w);
  
  CosSq<1><<<num_blocks, num_threads>>>( S->particleData.d_r, S->particleData.d_f, 
					 M->SB, M->SB->GetDevicePointer(), 
					 M->d_alist, M->d_aplist, M->d_epot_angle,
					 M->d_angles, num_threads, num_angles );

}

double AngleCosSq::GetPotentialEnergy(){

  unsigned num_angles = M->GetNumberOfAngles();

  CosSq<2><<<num_blocks, num_threads>>> ( S->particleData.d_r, S->particleData.d_f, 
					  M->SB, M->SB->GetDevicePointer(), 
					  M->d_alist, M->d_aplist, M->d_epot_angle,
					  M->d_angles, num_threads, num_angles );
  
  size_t nbytes = num_angles*sizeof(float);
  cudaMemcpy(M->h_epot_angle, M->d_epot_angle, nbytes, cudaMemcpyDeviceToHost);
  
  double epot = 0.0;
  for ( unsigned n=0; n<num_angles; n++ )  epot += (double)M->h_epot_angle[n];
    
  return epot;
}



template <int energy, class Simbox> 
__global__ void CosSq( float4 *r, float4 *f,  Simbox *simbox, float *simboxpointer, 
		       uint4 *alist, float2 *parameter, float *d_epot_angle, 
		       float *d_angles, int num_thread_per_block, int num_angles ){

  // First approach is to let one thread work on an angle
  float4 dr1, dr2;    
  float f1, f2;

  unsigned angle_index = blockIdx.x*num_thread_per_block + threadIdx.x;

  if ( angle_index < num_angles ){
    unsigned angle_type = alist[angle_index].w;

    unsigned a = alist[angle_index].x;
    unsigned b = alist[angle_index].y;
    unsigned c = alist[angle_index].z;

    float theta0 = parameter[angle_type].x;
    float ktheta = parameter[angle_type].y;

    const float cCon = cos((float) M_PI - theta0);

    // Load the simulation box in local memory to avoid bank conflicts. 
    float array[simulationBoxSize];
    simbox->loadMemory(array, simboxpointer);
  
    dr1 = simbox->calculateDistance(r[b], r[a], array);
    dr2 = simbox->calculateDistance(r[c], r[b], array);
  
    float c11 = dr1.w;
    float c12 = dr1.x*dr2.x + dr1.y*dr2.y + dr1.z*dr2.z;
    float c22 = dr2.w;

    float icD = 1.f/sqrtf(c11*c22);
    float cc = c12*icD;  

    d_angles[angle_index] = (float) M_PI - acos(cc); 
    float epot = 0.5f*ktheta*(cc-cCon)*(cc-cCon);

    if ( energy == 2 ){
      d_epot_angle[angle_index] = 0.5f*ktheta*(cc-cCon)*(cc-cCon); 
    }
    else if ( energy == 1) {

      AngleAtomicFloatAdd(&(f[a].w), epot);
 
      float ff = -ktheta*(cc - cCon);
      float c1 = c12/c11;
      float c2 = c12/c22;
    
      f1 = ff*(c1*dr1.x - dr2.x)*icD;
      f2 = ff*(dr1.x - c2*dr2.x)*icD;
    
      AngleAtomicFloatAdd(&(f[a].x), f1);
      AngleAtomicFloatAdd(&(f[b].x), -f1-f2);
      AngleAtomicFloatAdd(&(f[c].x), f2);
    
    
      f1 = ff*(c1*dr1.y - dr2.y)*icD;
      f2 = ff*(dr1.y - c2*dr2.y)*icD;
    
      AngleAtomicFloatAdd(&(f[a].y), f1);
      AngleAtomicFloatAdd(&(f[b].y), -f1-f2);
      AngleAtomicFloatAdd(&(f[c].y), f2);
    
      f1 = ff*(c1*dr1.z - dr2.z)*icD;
      f2 = ff*(dr1.z - c2*dr2.z)*icD;
    
      AngleAtomicFloatAdd(&(f[a].z), f1);
      AngleAtomicFloatAdd(&(f[b].z), -f1-f2);
      AngleAtomicFloatAdd(&(f[c].z), f2);
    }
  }

}
