#include <zlib.h>
#include <cstdio>
#include <cerrno>
#include <iostream>
#include <vector>
#include "rumd/RUMD_Error.h"

#include "rumd/ConfigurationWriterReader.h"
#include "rumd/Sample.h"
#include "rumd/Potential.h"
#include "rumd/PairPotential.h"
#include "rumd/IntegratorNVT.h"

void ConfigurationWriterReader::Write(Sample* S, const std::string& filename, const std::string& mode) {
  // meta-data	such as the box-length, which has meaning independent of
  // the file-format, is taken from sample class itself. We do not check
  // the passed metaData variable for consistency (ConfigurationMetaData has 
  // fields for such data only for reading in configurations)
  
  float dt = 0.;
  float thermostatState = 0;
  if(S->itg) dt = S->itg->GetTimeStep();
  IntegratorNVT* testItg = dynamic_cast<IntegratorNVT*>(S->itg);
  if(testItg){
    thermostatState = testItg->GetThermostatState();
  }
  
  int precision = metaData.precision;

  int molIdx = 0; //local, temporary---Trond will connect to real stuff
  int numMolecules = 0; // likewise
  
  gzFile gp;
  std::string compressedFilename(filename);
  int len = compressedFilename.size();
  // add .gz if not already present
  if(len < 3 || compressedFilename.substr(len-3,3).compare(".gz"))
    compressedFilename.append(".gz");

  // the mode is append if this is a trajectory file and the index is nonzero  
  gp = gzopen(compressedFilename.c_str(),mode.c_str());

  S->CopySimulationDataFromDevice();
  ParticleData& particleData = S->particleData;
  float4* h_r = particleData.h_r;
  float4* h_v = particleData.h_v;
  float4* h_im = particleData.h_im;
  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w; 
  unsigned int* h_Type = particleData.h_Type;

  unsigned int N = S->GetNumberOfParticles(); 
  // first line, number of particles
  gzprintf(gp, "%u\n", N); 
  
  // second line, "comment-line", containing meta-data
  float4 L = S->GetSimulationBox()->GetSimulationBox();
  
  gzprintf(gp, "ioformat=%d dt=%.*f", rumd_conf_ioformat, precision, dt);
  if(metaData.has_logLin)
    gzprintf(gp," timeStepIndex=%u logLin=%u,%u,%u,%u,%u", metaData.logLin.nextTimeStep, metaData.logLin.block, metaData.logLin.base,metaData.logLin.index, metaData.logLin.maxIndex, metaData.logLin.maxInterval);

  unsigned int numTypes = S->numTypes;
  gzprintf(gp, " boxLengths=%.*f,%.*f,%.*f numTypes=%d Nose-Hoover-Ps=%.*f Barostat-Pv=%.*f", precision, L.x, precision, L.y, precision, 
	  L.z, numTypes, precision, thermostatState, precision, 0.f); 
    

  float* massOfType = S->massOfType;
  gzprintf(gp, " mass=%.*f",precision,  massOfType[0]);
  for(unsigned int tdx = 1; tdx < numTypes; ++tdx)
    gzprintf(gp, ",%.*f", precision, massOfType[tdx]);
  
  if(metaData.has_moleculeIndices)
	gzprintf(gp, " numMolecules=%d", numMolecules);

  gzprintf(gp, " columns=type,x,y,z");
  if(metaData.has_images) gzprintf(gp, ",imx,imy,imz");
  if(metaData.has_velocities) gzprintf(gp, ",vx,vy,vz");
  if(metaData.has_forces) gzprintf(gp, ",fx,fy,fz");
  if(metaData.has_pot_energies) gzprintf(gp, ",pe");
  if(metaData.has_virials) gzprintf(gp, ",vir");
  if(metaData.has_moleculeIndices) gzprintf(gp, ",molIdx");
  gzprintf(gp, "\n");
  

  for ( unsigned int i = 0; i < N; i++ ){
    gzprintf(gp, "%d %.*f %.*f %.*f", h_Type[i], precision, h_r[i].x, 
	     precision, h_r[i].y, precision, h_r[i].z);
    if(metaData.has_images)
      gzprintf(gp, " %.0f %.0f %.0f", h_im[i].x, h_im[i].y, h_im[i].z);
    if(metaData.has_velocities)
      gzprintf(gp, " %.*f %.*f %.*f", precision, h_v[i].x, 
	       precision, h_v[i].y, precision, h_v[i].z);
    if(metaData.has_forces)
      gzprintf(gp, " %.*f %.*f %.*f", precision, h_f[i].x, 
	       precision, h_f[i].y, precision, h_f[i].z);
    if(metaData.has_pot_energies)
      gzprintf(gp, " %.*f", precision, h_f[i].w);
    if(metaData.has_virials)
      gzprintf(gp, " %.*f", precision, h_w[i].w/2.);
    if(metaData.has_moleculeIndices)
      gzprintf(gp, " %d", molIdx);
    gzprintf(gp,"\n");
    
    // Some consistency check for the neighborList.
    std::vector<Potential*>* vec = S->GetPotentials();
    for ( std::vector<Potential*>::iterator itr = vec->begin(); itr < vec->end(); itr++ ){
      PairPotential* testPairPotential = dynamic_cast<PairPotential*>(*itr);

      if( testPairPotential && h_im[i].w > 0.250001*testPairPotential->GetNbListSkin()*testPairPotential->GetNbListSkin() )
	throw( RUMD_Error("ConfigurationWriterReader","Write","Particles have moved more than half the neighborList skin. Check your starting configuration.") );
    }
  }
  
  gzclose(gp); 
}

void ConfigurationWriterReader::Read(Sample* S, const std::string& filename) {
  char strWorkspace[256];
  char *ok; // for checking success of gzgets
  unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
  unsigned int tdx;
  int nItems;
  unsigned int set_num_part = 0;
  float4 zero = { 0.f, 0.f, 0.f, 0.f };
    
  gzFile gp;
  gp = gzopen(filename.c_str(),"r");
  if (!gp)
    throw RUMD_Error("ConfigurationWriterReader","Read", std::string("\nError opening file ") + filename + ": " + strerror(errno));

  // read the first line
  ok = gzgets(gp, strWorkspace, maxLen);
  if(!ok) 
    throw RUMD_Error("ConfigurationWriterReader","Read", std::string("\nError reading first line of file ")+filename);

  // read number of particles, do associated memory alloc
  nItems = sscanf(strWorkspace, "%u", &set_num_part);
  if(nItems != 1)
    throw RUMD_Error("ConfigurationWriterReader","Read", std::string("\nError reading number of particles from file ")+filename);

  assert(set_num_part > 0);   
  S->SetNumberOfParticles(set_num_part);

  ParticleData& particleData = S->particleData;
  float4* h_r = particleData.h_r;
  float4* h_v = particleData.h_v;
  float4* h_im = particleData.h_im;
  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w; 
  unsigned int* h_Type = particleData.h_Type;

  // read the comment line
  ok = gzgets(gp, strWorkspace, maxLen);
  if(!ok || strlen(strWorkspace) == maxLen - 1)
    throw RUMD_Error("ConfigurationWriterReader","Read",std::string("\nError reading comment line in file ")+filename);
  else {
    // replace final newline with a zero
    int lenCmtLine = strlen(strWorkspace);
    assert(strcmp(&(strWorkspace[lenCmtLine-1]),"\n") == 0);
    strWorkspace[lenCmtLine-1] = 0;
  }
    
  // initialize metaData with known parameters (recall not all parameters need to be
  // given on the comment line)
  metaData.Ps = 0.0; // XXX ????? need to initialize it with something if not present in the file

  metaData.rumd_conf_ioformat = rumd_conf_ioformat;
  for(tdx = 0; tdx < MaxNumTypes; ++tdx)
    S->massOfType[tdx] = 1.0; // default value
  
  // read data from comment line into metaData
  metaData.ReadMetaData(strWorkspace, verbose);
  S->numTypes = metaData.numTypes;
  
  // Allocate SimulationBox.
  float boxDetails[7] = { 0, metaData.L[0], metaData.L[1], metaData.L[2], 0, 0, 0 };
  
  if(S->simulationBox)
    S->simulationBox->SetBoxDetails(boxDetails);
    
  assert(metaData.rumd_conf_ioformat <= rumd_conf_ioformat);
  // so far we have nothing to handle cases where rumd_conf_ioformat of the file
  // being read is different from the current rumd_conf_ioformat

  // masses, Ps and Pv are allowed to be determined by the comment line
  for(tdx = 0; tdx < S->numTypes; ++tdx)
    S->massOfType[tdx] = metaData.massOfType[tdx];
  
  IntegratorNVT* testItg = 0;
  if(S->itg) testItg = dynamic_cast<IntegratorNVT*>(S->itg);

  if(testItg)
    testItg->SetThermostatState(metaData.Ps);    
  else if(metaData.Ps) {
    IntegratorMetaData* itg_md = new IntegratorMetaData("IntegratorNVT");
    itg_md->AddParameter(metaData.Ps);
    S->SetIntegratorMetaData(itg_md);
    if(verbose)
      std::cout << "Thermostat-Ps variable from configuration file has been stored in Sample via IntegratorMetaData" << std::endl;
  }

  int numMolecules = metaData.numMolecules; // local.
  
  // this is required for the memcpy trick with Types
  assert(sizeof(int) == sizeof(float));
  //bool warnMissingPotential = false;
  std::vector<unsigned int> numberOfType(S->numTypes);
 
  // read the per-particle
  for ( unsigned int i = 0; i < S->GetNumberOfParticles(); i++ ){
    ok = gzgets(gp, strWorkspace, maxLen);
    if(!ok || strlen(strWorkspace) == maxLen - 1) {
      std::ostringstream errorStr;
      errorStr << std::string("\nError reading line ") << i << " from configuration file " << filename;
      throw RUMD_Error("ConfigurationWriterReader","Read",errorStr.str());
    }

    int offset = 0;
    int partial_offset = 0;
    
    // type and positions
    sscanf(strWorkspace+offset, "%d %f %f %f%n", &h_Type[i], &h_r[i].x, &h_r[i].y, &h_r[i].z, &partial_offset);
    offset += partial_offset;
    numberOfType[h_Type[i]] += 1;
    // set mass
    if (h_Type[i] >= S->numTypes) {
      printf("%d %f %f %f %d\n", h_Type[i], h_r[i].x, h_r[i].y, h_r[i].z, S->numTypes );
      assert(h_Type[i] < S->numTypes);	    
    }
    // For sending to the device, store type as a float instead of uint

    memcpy((void*)&(h_r[i].w), (void*)&(h_Type[i]),sizeof(int));

    h_v[i].w = 1.0 / S->massOfType[h_Type[i]];
    h_im[i] = zero; // initialize images
    if(metaData.has_images) {
      sscanf(strWorkspace+offset, "%f %f %f%n", &h_im[i].x, &h_im[i].y, &h_im[i].z, &partial_offset);
      offset += partial_offset;
    }
    if(metaData.has_velocities)
      {
	sscanf(strWorkspace+offset, "%f %f %f%n", &h_v[i].x, &h_v[i].y, &h_v[i].z, &partial_offset);
	offset += partial_offset;
      }	
    if(metaData.has_forces) {
      sscanf(strWorkspace+offset, "%f %f %f%n", &h_f[i].x, &h_f[i].y, &h_f[i].z, &partial_offset);
      offset += partial_offset;
    }
    if(metaData.has_pot_energies) {
      sscanf(strWorkspace+offset, "%f%n", &h_f[i].w, &partial_offset);
      offset += partial_offset;
    }
    if(metaData.has_virials) {
      sscanf(strWorkspace+offset, "%f%n",&h_w[i].w, &partial_offset);
      offset += partial_offset;
      h_w[i].w *= 2.;
    }
    if(metaData.has_moleculeIndices) {
      int dummy; 
      sscanf(strWorkspace+offset, "%d%n", &dummy, &partial_offset);
      offset += partial_offset;
      assert(0 <= dummy && dummy < numMolecules);
    }
  }
  unsigned int check_type_sum = 0;
  for (unsigned int tdx=0;tdx < S->numTypes; ++tdx) {
    check_type_sum += numberOfType[tdx];
    S->numberOfType[tdx] = numberOfType[tdx];
  }
  for (unsigned int tdx=S->numTypes;tdx < MaxNumTypes; ++tdx)
    S->numberOfType[tdx] = 0;
  assert ( check_type_sum  == S->GetNumberOfParticles() );

  gzclose(gp);

  // Check if the type number of particles is allowed. Temp. solution.
  for( unsigned i=0; i < S->GetNumberOfParticles(); i++ ){
    if( unsigned(h_r[i].w) > 31 ) 
      throw( RUMD_Error("ConfigurationWriterReader","Read", "The particle type number in start.xyz is larger than 31" ) );
  }
  
  S->CopySimulationDataToDevice();    
}
